from importlib_metadata import version
from onediff.utils import logger


def get_support_message():
    message = f""" OneDiff Enterprise Edition features can't be used, please refer to here for help: https://github.com/siliconflow/onediff?tab=readme-ov-file#community--support """
    return message


def is_quantization_enabled():
    import oneflow

    if version("oneflow") < "0.9.1":
        logger.warning(
            f"onediff_comfy_nodes requires oneflow>=0.9.1 to run, {get_support_message()}"
        )
        return False
    try:
        import onediff_quant
    except ImportError as e:
        logger.warning(
            f"Failed to import onediff_quant, Error message: {e}, {get_support_message()}"
        )
        return False
    return hasattr(oneflow._C, "dynamic_quantization")


def is_community_version():
    import oneflow.sysconfig

    return not oneflow.sysconfig.with_enterprise()



from functools import wraps
import oneflow as flow
import time
import inspect
from onediff.utils import logger

__all__ = ["cost_cnt", "cost_time"]


class cost_cnt:
    def __init__(self, debug=False, message="\t"):
        self.debug = debug
        self.message = message

    def __enter__(self):
        if not self.debug:
            return
        flow._oneflow_internal.eager.Sync()
        before_used = flow._oneflow_internal.GetCUDAMemoryUsed()
        before_host_used = flow._oneflow_internal.GetCPUMemoryUsed()
        logger.debug(f"====> {self.message} try to run...")
        logger.debug(f"{self.message} cuda mem before {before_used} MB")
        logger.debug(f"{self.message} host mem before {before_host_used} MB")
        self.before_used = before_used
        self.before_host_used = before_host_used
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.debug:
            return
        flow._oneflow_internal.eager.Sync()
        end_time = time.time()
        after_used = flow._oneflow_internal.GetCUDAMemoryUsed()
        after_host_used = flow._oneflow_internal.GetCPUMemoryUsed()
        logger.debug(f"{self.message} run time {end_time - self.start_time} seconds")
        logger.debug(f"{self.message} cuda mem after {after_used} MB")
        logger.debug(f"{self.message} cuda mem diff {after_used - self.before_used} MB")
        logger.debug(f"{self.message} host mem after {after_host_used} MB")
        logger.debug(
            f"{self.message} host mem diff {after_host_used - self.before_host_used} MB"
        )
        logger.debug(f"<==== {self.message} finish run.")

    def __call__(self, func):
        @wraps(func)
        def clocked(*args, **kwargs):
            if not self.debug:
                return func(*args, **kwargs)
            module = inspect.getmodule(func)
            logger.debug(
                f"==> function {module.__name__}.{func.__name__}  try to run..."
            )
            flow._oneflow_internal.eager.Sync()

            before_used = flow._oneflow_internal.GetCUDAMemoryUsed()
            logger.debug(f"{func.__name__} cuda mem before {before_used} MB")

            before_host_used = flow._oneflow_internal.GetCPUMemoryUsed()
            logger.debug(f"{func.__name__} host mem before {before_host_used} MB")

            start_time = time.time()
            out = func(*args, **kwargs)
            flow._oneflow_internal.eager.Sync()
            end_time = time.time()

            logger.debug(f"{func.__name__} run time {end_time - start_time} seconds")

            after_used = flow._oneflow_internal.GetCUDAMemoryUsed()
            logger.debug(f"{func.__name__} cuda mem after {after_used} MB")

            logger.debug(f"{func.__name__} cuda mem diff {after_used - before_used} MB")
            after_host_used = flow._oneflow_internal.GetCPUMemoryUsed()
            logger.debug(f"{func.__name__} host mem after {after_host_used} MB")
            logger.debug(
                f"{func.__name__} host mem diff {after_host_used - before_host_used} MB"
            )

            logger.debug(f"<== function {func.__name__} finish run.")
            logger.debug("")
            return out

        return clocked


class cost_time:
    """
    simple cost time code ranges using a context manager or a decorator.
    """

    def __init__(self, debug=False, message="\t"):
        self.debug = debug
        self.message = message

    def __enter__(self):
        if not self.debug:
            return
        flow._oneflow_internal.eager.Sync()
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.debug:
            return
        flow._oneflow_internal.eager.Sync()
        end_time = time.time()
        logger.debug(f"{self.message} run time {end_time - self.start_time} seconds")

    def __call__(self, func):
        @wraps(func)
        def clocked(*args, **kwargs):
            if not self.debug:
                return func(*args, **kwargs)
            module = inspect.getmodule(func)
            logger.debug(
                f"==> function {module.__name__}.{func.__name__}  try to run..."
            )
            flow._oneflow_internal.eager.Sync()
            start_time = time.time()
            out = func(*args, **kwargs)
            flow._oneflow_internal.eager.Sync()
            end_time = time.time()
            logger.debug(f"{func.__name__} run time {end_time - start_time} seconds")
            logger.debug(f"<== function {func.__name__} finish run.")
            logger.debug("")
            return out

        return clocked



import torch

from .deployable_module import DeployableModule

_DEFAULT_BACKEND = "oneflow"


def compile(
    torch_module: torch.nn.Module, *, backend=_DEFAULT_BACKEND, options=None
) -> DeployableModule:
    from .registry import lookup_backend

    backend = lookup_backend(backend)
    model = backend(torch_module, options=options)
    return model


def oneflow_compile(torch_module: torch.nn.Module, *, options=None) -> DeployableModule:
    return compile(torch_module, backend="oneflow", options=options)



from diffusers import AutoPipelineForText2Image
from onediff.quantization.quantize_pipeline import QuantPipeline
import torch
import argparse


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--floatting_model_path", default="runwayml/stable-diffusion-v1-5"
    )
    parser.add_argument(
        "--prompt", default="a photo of an astronaut riding a horse on mars"
    )
    parser.add_argument("--height", type=int, default=1024)
    parser.add_argument("--width", type=int, default=1024)
    parser.add_argument("--num_inference_steps", type=int, default=30)
    parser.add_argument("--conv_compute_density_threshold", type=int, default=900)
    parser.add_argument("--linear_compute_density_threshold", type=int, default=300)
    parser.add_argument("--conv_ssim_threshold", type=float, default=0.985)
    parser.add_argument("--linear_ssim_threshold", type=float, default=0.991)
    parser.add_argument("--save_as_float", type=bool, default=False)
    parser.add_argument("--cache_dir", default="./run_sd-v1-5")
    parser.add_argument("--quantized_model", default="./quantized_model")
    return parser.parse_args()


args = parse_args()

pipe = QuantPipeline.from_pretrained(
    AutoPipelineForText2Image,
    args.floatting_model_path,
    torch_dtype=torch.float16,
    variant="fp16",
    use_safetensors=True,
)
pipe.to("cuda")

pipe_kwargs = dict(
    prompt=args.prompt,
    height=args.height,
    width=args.width,
    num_inference_steps=args.num_inference_steps,
)

pipe.quantize(
    **pipe_kwargs,
    conv_compute_density_threshold=args.conv_compute_density_threshold,
    linear_compute_density_threshold=args.linear_compute_density_threshold,
    conv_ssim_threshold=args.conv_ssim_threshold,
    linear_ssim_threshold=args.linear_ssim_threshold,
    save_as_float=args.save_as_float,
    plot_calibrate_info=False,
    cache_dir=args.cache_dir
)

pipe.save_quantized(args.quantized_model, safe_serialization=True)



import torch


def should_patch_torch_module(torch_module):
    # https://github.com/comfyanonymous/ComfyUI/blob/master/comfy/ops.py#L132-L134
    supported_types = (
        torch.nn.ConvTranspose2d,
        torch.nn.GroupNorm,
        torch.nn.Conv3d,
        torch.nn.Conv2d,
        torch.nn.Linear,
    )
    full_cls_name = (
        type(torch_module).__module__ + "." + type(torch_module).__qualname__
    )
    return (
        isinstance(torch_module, supported_types)
        or full_cls_name.endswith("controlnet.ControlLoraOps.Linear")
        or full_cls_name.endswith("controlnet.ControlLoraOps.Conv2d")
    )


def apply_comfy_settings(torch_module, flow_module):
    keys = [
        "comfy_cast_inputs",
        "comfy_cast_weights",
        "weight_function",
        "bias_function",
    ]
    for k in keys:
        setattr(flow_module, k, getattr(torch_module, k, None))


class PatchForComfy:
    def __init__(self, oneflow_model):
        self.oneflow_model = oneflow_model

    def patch(
        self,
        torch_module,
        should_patch: callable = None,
        apply_settings: callable = None,
    ):
        if should_patch(torch_module):
            apply_settings(torch_module, self.oneflow_model)

    def __call__(self, torch_module):
        self.patch(torch_module, should_patch_torch_module, apply_comfy_settings)



import torch
import oneflow as flow
from oneflow.framework.args_tree import ArgsTree
from onediff.utils import logger


def input_output_processor(func):
    def process_input(*args, **kwargs):
        def input_fn(value):
            if isinstance(value, torch.Tensor):
                # TODO: https://github.com/siliconflow/sd-team/issues/109
                return flow.utils.tensor.from_torch(value.contiguous())
            else:
                return value

        args_tree = ArgsTree((args, kwargs), False, tensor_type=torch.Tensor)
        input_count = len(
            [v for v in args_tree.iter_nodes() if isinstance(v, torch.Tensor)]
        )
        out = args_tree.map_leaf(input_fn)
        mapped_args = out[0]
        mapped_kwargs = out[1]
        return mapped_args, mapped_kwargs, input_count

    def process_output(output):
        def output_fn(value):
            if isinstance(value, flow.Tensor):
                return flow.utils.tensor.to_torch(value)
            else:
                return value

        out_tree = ArgsTree((output, None), False)
        out = out_tree.map_leaf(output_fn)
        return out[0]

    def wrapper(self: "DeployableModule", *args, **kwargs):
        mapped_args, mapped_kwargs, input_count = process_input(*args, **kwargs)
        if (
            self._deployable_module_options.use_graph
            and self._deployable_module_dpl_graph is not None
        ):
            count = len(self._deployable_module_dpl_graph._input_op_names)
            if count != input_count:
                logger.warning(
                    f"Module {type(self._deployable_module_model.oneflow_module)} input tensor count changed from {count} to {input_count}, will compile again."
                )
                self._deployable_module_dpl_graph = None
                self._load_graph_first_run = True

        output = func(self, *mapped_args, **mapped_kwargs)
        return process_output(output)

    return wrapper
